// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Main program, simply calls D_DoomMain high level loop.
//
//-----------------------------------------------------------------------------

static const char
rcsid[] = "$Id: i_main.c,v 1.4 1997/02/03 22:45:10 b1 Exp $";



#include "doomdef.h"

#include "m_argv.h"
#include "d_main.h"

#include "mb_interface.h"
#include "xparameters.h"
#include "xintc.h"
#include "xps2.h"

#include "../stage2boot/gfxlib.c"

void Cache ( void )
{
    unsigned i ;

    microblaze_disable_icache () ;
    microblaze_disable_dcache () ;
    for (i = 0; i < 65536; i += 4)
    {
        asm volatile ( 
                "addik r3, %0, 0\n"
                "wic r3, r0\n"
                "nop\n"
                "addik r3, %0, 0\n"
                "wdc r3, r0\n"
                "nop\n" : : "r"(i) : "r3" ) ;
    }
    microblaze_enable_icache () ;
    microblaze_enable_dcache () ;
}

typedef void (* Target_Fn) ( void ) ;

static XIntc intc;
static XPs2 ps2;
static unsigned ps2_extend;

static int translate_key(unsigned code)
{
    int rc =0;
    switch(code)
    {
      case 0x16b:	rc = KEY_LEFTARROW;	break;
      case 0x174:	rc = KEY_RIGHTARROW;	break;
      case 0x172:	rc = KEY_DOWNARROW;	break;
      case 0x175:	rc = KEY_UPARROW;	break;
      case 0x29:	rc = ' ';	        break;
      case 0x76:	rc = KEY_ESCAPE;	break;
      case 0x5a:	rc = KEY_ENTER;		break;
      case 0x0d:	rc = KEY_TAB;		break;
      case 0x05:	rc = KEY_F1;		break;
      case 0x06:	rc = KEY_F2;		break;
      case 0x04:	rc = KEY_F3;		break;
      case 0x0c:	rc = KEY_F4;		break;
      case 0x03:	rc = KEY_F5;		break;
      case 0x0b:	rc = KEY_F6;		break;
      case 0x83:	rc = KEY_F7;		break;
      case 0x0a:	rc = KEY_F8;		break;
      case 0x01:	rc = KEY_F9;		break;
      case 0x07:	rc = KEY_F12;		break;
      case 0x41:	rc = ',';		    break; // strafe left
      case 0x49:	rc = '.';		    break; // strafe right
    
      case 0x66:
      case 0x171:	rc = KEY_BACKSPACE;	break;

      case 0x45:    rc = '0'; break;
      case 0x16:    rc = '1'; break;
      case 0x1e:    rc = '2'; break;
      case 0x26:    rc = '3'; break;
      case 0x25:    rc = '4'; break;
      case 0x2e:    rc = '5'; break;
      case 0x36:    rc = '6'; break;
      case 0x3d:    rc = '7'; break;
      case 0x3e:    rc = '8'; break;
      case 0x46:    rc = '9'; break;
      case 0x1c:    rc = 'a'; break;
      case 0x32:    rc = 'b'; break;
      case 0x21:    rc = 'c'; break;
      case 0x23:    rc = 'd'; break;
      case 0x24:    rc = 'e'; break;
      case 0x2b:    rc = 'f'; break;
      case 0x34:    rc = 'g'; break;
      case 0x33:    rc = 'h'; break;
      case 0x43:    rc = 'i'; break;
      case 0x3b:    rc = 'j'; break;
      case 0x42:    rc = 'k'; break;
      case 0x4b:    rc = 'l'; break;
      case 0x3a:    rc = 'm'; break;
      case 0x31:    rc = 'n'; break;
      case 0x44:    rc = 'o'; break;
      case 0x4d:    rc = 'p'; break;
      case 0x15:    rc = 'q'; break;
      case 0x2d:    rc = 'r'; break;
      case 0x1b:    rc = 's'; break;
      case 0x2c:    rc = 't'; break;
      case 0x3c:    rc = 'u'; break;
      case 0x2a:    rc = 'v'; break;
      case 0x1d:    rc = 'w'; break;
      case 0x22:    rc = 'x'; break;
      case 0x35:    rc = 'y'; break;
      case 0x1a:    rc = 'z'; break;


      case 0x55:
      case 0x79:	rc = KEY_EQUALS;	break;

      case 0x7b:
      case 0x4e:	rc = KEY_MINUS;		break;

      case 0x12:
      case 0x59:    rc = KEY_RSHIFT; break;
    
      case 0x14:
      case 0x114:   rc = KEY_RCTRL; break;
    
      case 0x11:
      case 0x111:   rc = KEY_RALT; break;
    
      default:
        break;
    }

    return rc;
}

#define MAX_EVENTS      100
static volatile int     kb_num_events = 0;
static volatile event_t kb_event_queue[ MAX_EVENTS ];

static void
Keyboard_ISR (void *CallBackRef, u32 IntrMask, u32 NumBytes)
{
    u8          key;
    int         tkey;

    if (!(IntrMask & XPS2_IPIXR_RX_FULL)) {
        return;
    }
    while (XPs2_IsRxFull(XPAR_PS2_0_BASEADDR)) {
        key = XPs2_RecvByte(XPAR_PS2_0_BASEADDR);
        
        if (key == 0xe0) {
            ps2_extend |= 0x100;
        } else if (key == 0xf0) {
            ps2_extend |= 0x200;
        } else {
            ps2_extend |= key;
            tkey = translate_key(ps2_extend & 0x1ff);
            if ((tkey != 0)
            && (kb_num_events < MAX_EVENTS)) {
                event_t * event = &kb_event_queue[kb_num_events];
                event->data1 = tkey;
                if (ps2_extend & 0x200) {
                    event->type = ev_keyup;
                } else {
                    event->type = ev_keydown;
                }
                kb_num_events ++;
            }
            ps2_extend = 0;
        }
    }
}

void I_GetEvent()
{
    int i = 0;

    microblaze_disable_interrupts(); // Critical section
    for (i = 0; i < kb_num_events; i++) {
        D_PostEvent(&kb_event_queue[i]);
    }
    kb_num_events = 0;
    microblaze_enable_interrupts();
}


int
main
( int		argc,
  char**	argv ) 
{ 
    XPs2_Config *   ps2cfg = XPs2_LookupConfig (XPAR_PS2_0_DEVICE_ID);
    int             rc;

    myargc = argc; 
    myargv = argv; 

    printf("Arrived in Doom\n");
    Stop_Polling(); 
    memcpy ( (void *) 0x8 , 
            (void *) ( (unsigned) sdram + 0x8 ) , 0x18 ) ; /* Nasty! */
    printf("Block RAM patched\n");

    Cache();
    if (ps2cfg == NULL) {
        printf("No keyboard driver available - crash.\n");
        return 0;
    }
    ps2_extend = 0;
    rc = XPs2_CfgInitialize(&ps2, ps2cfg, XPAR_PS2_0_BASEADDR);
    if (rc != 0) {
        printf("XPs2_CfgInitialize rc = %d - crash.\n", rc);
        return 0;
    }
    printf("Initialise 1\n");
    microblaze_register_handler (
        (XInterruptHandler) XIntc_InterruptHandler,
        & intc );
    rc = XIntc_Initialize(&intc, XPAR_XPS_INTC_0_DEVICE_ID);
    if ((rc != XST_SUCCESS) && (rc != XST_DEVICE_IS_STARTED)) {
        printf("XIntc_Initialize rc = %d - crash.\n", rc);
        return 0;
    }
    printf("Initialise 2\n");
    rc = XIntc_Connect( &intc,
            XPAR_INTC_0_PS2_0_VEC_ID,
            (XInterruptHandler)XPs2_IntrHandler, &ps2);
    if (rc != XST_SUCCESS) {
        printf("XIntc_Connect(ps2) rc = %d - crash.\n", rc);
        return 0;
    }
    rc = XIntc_Start ( &intc, XIN_REAL_MODE);
    if (rc != XST_SUCCESS) {
        printf("XIntc_Start rc = %d - crash.\n", rc);
        return 0;
    }
    printf("Initialise 3\n");

    XPs2_SetHandler(&ps2, Keyboard_ISR, NULL);
    XPs2_IntrGlobalEnable(&ps2);
    XPs2_IntrEnable(&ps2, XPS2_IPIXR_RX_ALL);
    printf("Intc/PS2 setup complete\n");

    XIntc_Enable (&intc,
            XPAR_INTC_0_PS2_0_VEC_ID );
    microblaze_enable_interrupts();

#if 0
    printf("Interrupts enabled.\nPre-splash.\n");
    {
        int     i, x, y, c;

        char * splash = malloc ( 64000 ) ;
        for ( i = 0 ; i < 64000 ; i ++ )
        {
            y = i / 320;
            x = i % 320;
            x -= 20;
            y -= 10;
            splash [ i ] = 0; 
            if (( x >= 0 ) && ( y >= 0 ) && ( x < 256 ))
            {
                c = (( x / 16 ) % 16 ) + (( y / 11 ) * 16 );
                if ( c < 256 )
                {
                    splash [ i ] = c ; 
                }
            }
        }
        vga [ 0 ] = (unsigned) splash;
        while (kb_num_events == 0) {}
        memset (splash, 0, 64000);
        //vga [ 0 ] = 0;
        //free ( splash ) ;
    }
#endif
    
    printf("Start D_DoomMain()\n");
    D_DoomMain (); 

    //while (1) { I_GetEvent(); }

    return 0;
} 


void exit(int status)
{
    Target_Fn target_code = (Target_Fn) bootloader;
    printf("exit() called\n");
    fflush(stdout);
    vga [ 0 ] = 0; // blackout
    microblaze_disable_interrupts(); 
    target_code () ;
    while (1) {}
}



